/* Copyright 2008-2014 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
begin_html
<b><a href="index.html">Up</a></b>
<ul>
<li><a href="qubfast.h.html">qubfast.h</a></li>
<li><a href="qub_markov.h.html">qub_markov.h</a></li>
<li><a href="qub_markov.cpp.html">qub_markov.cpp</a></li>
<li><a href="ublas_plus.h.html">ublas_plus.h</a></li>
<li><a href="ublas_matrixutil.h.html">ublas_matrixutil.h</a></li>
<li><a href="ublas_matrixutil.cpp.html">ublas_matrixutil.cpp</a></li>
<li><a href="matrix.h.html">matrix.h</a></li>
</ul>
end_html
*/

#ifdef EMSCRIPTEN
  #include <emscripten.h>
  #define QUBFAST_API EMSCRIPTEN_KEEPALIVE
  #ifdef __GNUC__
  #define QUBFAST_VAR_NOT_USED __attribute__ ((unused))
  #else
  #define QUBFAST_VAR_NOT_USED
  #endif
#else
  #include <omp.h>
#endif

#ifdef _WIN32
  #include <windows.h>
  #include <time.h>
  #include <process.h>
  #define sleep Sleep
  #define getpid _getpid
  #define M_PI 3.141593653589793
#else
  #include <stdlib.h>
  #include <unistd.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <math.h>
#include <string.h>
#include <vector>

#include <boost/tuple/tuple.hpp>
using boost::tuple;

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/exponential_distribution.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/variate_generator.hpp>

#include "qubx_simulate.h"
#include "qubx_model_storage.h"
#include "max_ll_util.h"
#include "ublas_plus.h"
#include "ublas_matrixutil.h"

// boost::random: exponential, normal, and uniform distributions
typedef boost::mt19937 base_generator_type;
typedef boost::uniform_int<> seed_distribution_type;
typedef boost::variate_generator<base_generator_type&, seed_distribution_type> seed_gen_type;
typedef boost::uniform_real<> uni_distribution_type;
typedef boost::variate_generator<base_generator_type&, uni_distribution_type> uni_gen_type;
typedef boost::exponential_distribution<> expo_distribution_type;
typedef boost::variate_generator<base_generator_type&, expo_distribution_type> expo_gen_type;
typedef boost::normal_distribution<> norm_distribution_type;
typedef boost::variate_generator<base_generator_type&, norm_distribution_type> norm_gen_type;

// library lifetime to avoid duplicate seeds
//   func-lifetime generators with passed-in seed
base_generator_type g_randgen_qubxsim( static_cast<boost::mt19937::result_type>( clock() + (int)getpid() ) );


#ifdef _WIN32
  #include "../qubfast/gettimeofday.h"
#else
  #include <sys/time.h>
#endif
void print_timestamp(QUBFAST_VAR_NOT_USED
		     const char *lbl)
{
  //struct timeval tm;
  //gettimeofday(&tm, NULL);
  //cerr << "--- " << tm.tv_sec << " : " << (tm.tv_usec / 1000) << "." << (tm.tv_usec % 1000) << " --- " << lbl << endl;
}

long timedelta_ms(struct timeval *a, struct timeval *b)
{
  return ( (b->tv_sec - a->tv_sec)*1000 + (b->tv_usec - a->tv_usec)/1000 );
}

long timedelta_us(struct timeval *a, struct timeval *b)
{
  return ( (b->tv_sec - a->tv_sec)*1000000 + (b->tv_usec - a->tv_usec) );
}

class timediffs
{
public:
  struct timeval last_tm;
  std::vector<long> diffs;
  int i, N;
  timediffs(int N_)
    : diffs(N_), N(N_)
  { reset(); }
  void reset() {
    gettimeofday(&last_tm, NULL);
    i = 0;
  }
  void diff(const std::string& label) {
    if ( i == N ) return;
    struct timeval tm;
    gettimeofday(&tm, NULL);
    diffs[i++] = timedelta_us(&last_tm, &tm);
    last_tm = tm;
    cerr << label << endl;
  }
};

std::ostream& operator<<(std::ostream& out, const timediffs& td)
{
  for (int i=0; i<td.i; ++i)
    out << td.diffs[i] << ", ";
  return out;
}


/*
Stimulus interval iterator

  Each signal is either constant or sampled; constants have sample count = -1.
  The iterator presents an array of values[nstimulus] at the current time,
  and the count of samples until it changes.

*/
class stim_intervals
{
private:
  int nsamp;
  int nstim;
  int *counts;
  float **stims;
  std::vector<int> stor_ixs;
  int *ixs;
  std::vector<int> stor_remain;
  int *remain;
  std::vector<double> stor_values;
public:
  double *values;
  int count;
  
  stim_intervals(int nsample, int nstimulus, int *stim_counts, float **stimuli)
    : nsamp( nsample ),
      nstim( nstimulus ),
      counts( stim_counts ),
      stims( stimuli ),
      stor_ixs( nstim ),
      ixs( nstim ? &(stor_ixs[0]) : NULL ),
      stor_remain( nstim ),
      remain( NULL ), // stay NULL in case no stim // remain( nstim ? &(stor_remain[0]) : NULL ),
      stor_values( nstim ),
      values( nstim ? &(stor_values[0]) : NULL ),
      count( 0 )
  {
    reset();
  }
  
  stim_intervals(const stim_intervals& other)
    : nstim( other.nstim ),
      counts( other.counts ),
      stims( other.stims ),
      stor_ixs( nstim ),
      ixs( nstim ? &(stor_ixs[0]) : NULL ),
      stor_remain( nstim ),
      remain( (nstim && other.remain) ? &(stor_remain[0]) : NULL ),
      stor_values( nstim ),
      values( nstim ? &(stor_values[0]) : NULL ),
      count( other.count )
  {
    if ( nstim ) {
      memcpy(ixs, other.ixs, nstim*sizeof(int));
      memcpy(remain, other.remain, nstim*sizeof(int));
      memcpy(values, other.values, nstim*sizeof(int));
    }
  }
  
  void operator=(const stim_intervals& other)
  {
    if (this == &other)
      return;
    bool resize = (nstim != other.nstim);
    nstim = other.nstim;
    counts = other.counts;
    stims = other.stims;
    if ( resize && nstim ) {
      stor_ixs.resize( nstim );
      ixs = &(stor_ixs[0]);
      stor_remain.resize( nstim );
      remain = other.remain ? &(stor_remain[0]) : NULL;
      stor_values.resize( nstim );
      values = &(stor_values[0]);
    }
  }

  bool operator==(const stim_intervals& other) const {
    if ( nstim != other.nstim || counts != other.counts )
      return false;
    for (int i=nstim-1; i>=0; --i)
      if ( ixs[i] != other.ixs[i] )
	return false;
    return true;
  }
  bool operator!=(const stim_intervals& other) const { return ! (*this == other); }

  void reset() {
    if ( nstim == 0 )
      return;
    bool any_vary = false;
    for (int i=0; i<nstim; ++i)
      if ( counts[i] > 0 )
	any_vary = true;
    count = 1;
    remain = &(stor_remain[0]);
    for (int i=0; i<nstim; ++i) {
      if ( counts[i] == 0 ) {
	ixs[i] = 0;
	values[i] = stims[i][0];
	remain[i] = nsamp;
	if ( any_vary )
	  ++remain[i]; // the first ++(*this) will subtract one
      } else {
	ixs[i] = -1;
	remain[i] = 1;
      }
    }
    if ( any_vary )
      ++(*this);
    else
      count = nsamp;
  }
  
  operator bool () const {
    return ( remain != NULL );
  }
  
  stim_intervals& operator++ ()
  {
    if ( remain == NULL )
      return *this;
    int min_remain = (1<<30);
    for (int i=0; i<nstim; ++i) {
      if ( remain[i] > count ) {
	ixs[i] += count;
	remain[i] -= count;
      } else {
	ixs[i] += count;
	if ( ixs[i] < counts[i] ) {
	  float val = stims[i][ixs[i]];
	  values[i] = val;
	  int j = ixs[i]+1;
	  while ( (j<counts[i]) && (stims[i][j] == val) )
	    ++j;
	  remain[i] = j - ixs[i];
	}
	else {
	  count = 0;
	  remain = NULL;
	  return *this;
	}
      }
      if ( remain[i] < min_remain )
	min_remain = remain[i];
    }
    count = min_remain;
    
    //cerr << count << " x ";
    //for (int i=1; i<nstim; ++i)
    //  cerr << values[i] << "\t";
    //cerr << endl;
    
    return *this;
  }
};


void qubx_simulate_setup_P0(int Nchannel, int Nstate, int *start_states, double *P0,
			    matrix<double>& K0, matrix<double>& K1, matrix<double>& K2, matrix<int>& L, matrix<int>& V, matrix<int>& P,
			    double *stim_values, uni_gen_type& uni_gen)
{
  // pick random start state, if it's undefined (< 0)
  if ( *start_states < 0 ) {
    // copy entry probabilities, or calculate equilibrium at initial stimulus
    matrix<double> _P0(Nstate, Nstate);
    if ( P0 ) {
      matrix_from_array(_P0, P0, 1, Nstate);
    } else {
      matrix<double> Q(Nstate, Nstate);
      BuildQ_p(Nstate, Q, K0, K1, K2, L, V, P, stim_values);
      QtoPe(Q, _P0);
    }
    double P0sum = 0.0;
    for (int i=0; i<Nstate; ++i)
      P0sum += _P0(0,i);
    if ( P0sum > 0.0 ) {
      for (int i=0; i<Nstate; ++i)
	_P0(0,i) /= P0sum;
    } else {
      for (int i=0; i<Nstate; ++i)
	_P0(0,i) = 1.0 / Nstate;
    }
  
    for (int i=0; i<Nchannel; ++i) {
      double rand = uni_gen(); // uniform in [0,1) since P0 sums to 1
      double tot = _P0(0,0);
      start_states[i] = 0;
      while ( (tot < rand) && (start_states[i] < Nstate) )
	tot += _P0(0, ++(start_states[i]));
    }
  }
}



class qubx_simulate_A_single_channel
{
public:
  int st, next_st;
  int rep_count;
  
  qubx_simulate_A_single_channel(int start_state)
    : st( start_state ), next_st( start_state ), rep_count(0)
  {}
  
  template<class UniRand>
  int next_sample(matrix<double>& A, UniRand& uni, int Nstate)
  // A = exp(Q*dt) modified for cumulative prob: A[i,j] = sum(k=0..j of A_orig[i,k]) in qubx_simulate_A_kinetics; returns this->st;
  {
    int s = next_st;
    if ( s != st )
      rep_count = 1;
    else
      ++rep_count;
    st = s;
    
    double staterand = uni();
    for (int i=0; i<Nstate; ++i) {
      if ( A(s,i) > staterand ) {
	next_st = i;
	break;
      }
    }
    return s;
  }
};


class qubx_simulate_A_kinetics
// Maintains A (and Q) matrix corresponding to particular stimulus values.
{
public:
  int Ns;
  matrix<double> &K0, &K1, &K2;
  matrix<int> &L, &V, &P;
  double sampling;
  matrix<double> Q, A;
  std::vector<double> Qeig_r, Qeig_i;
  matrix<double> Qevec, Qeinv, Atmp1, Atmp2;
  
  qubx_simulate_A_kinetics(int Nstate, matrix<double>& K0_, matrix<double>& K1_, matrix<double>& K2_, matrix<int>& L_, matrix<int>& V_, matrix<int>& P_, double sampling_)
    : Ns(Nstate), K0(K0_), K1(K1_), K2(K2_), L(L_), V(V_), P(P_), sampling(sampling_),
      Q(Nstate, Nstate), A(Nstate, Nstate),
      Qeig_r(Nstate), Qeig_i(Nstate),
      Qevec(Nstate, Nstate), Qeinv(Nstate, Nstate), Atmp1(Nstate, Nstate), Atmp2(Nstate, Nstate)
  {
    Atmp1.clear(); // required zero off-diagonal
  }
  void setup_A(double *stim_values)
  {
    BuildQ_p(Ns, Q, K0, K1, K2, L, V, P, stim_values);
    eigen_vals_vecs(Q, &Qeig_r[0], &Qeig_i[0], Qevec, Qeinv, cerr);
    expm_eig(&Qeig_r[0], Qevec, Qeinv, sampling, Atmp1, Atmp2, A);
    // change A to cumulative probability for speed in inner loop
    for (int i=0; i<Ns; ++i) {
      double tot = 0.0;
      for (int j=0; j<Ns; ++j) {
	tot += A(i, j);
	A(i, j) = tot;
      }
    }
  }
};



class qubx_simulate_Q_single_channel
{
public:
  int st, next_st;
  double dur;
  
  qubx_simulate_Q_single_channel(int start_state)
    : st( start_state ), next_st( start_state )
  {}
  
  template<class ExpoRand, class UniRand>
  int next_event(matrix<double>& Q, ExpoRand& expo, UniRand& uni, double maxdur, int nstate)
  // returns exit_state
  {
    st = next_st;
    double full_duration = expo() / (- Q(st, st));
    if ( full_duration > maxdur ) {
      dur = maxdur;
      // next_st = st;
    }
    else {
      dur = full_duration;
      double staterand = uni() * -Q(st, st);
      double tot = 0.0;
      for (int i=0; i<nstate; ++i) {
	if (i != st) {
	  tot += Q(st, i);
	  if ( tot > staterand ) {
	    next_st = i;
		break;
	  }
	}
      }
    }
    return next_st;
  }
};


int qubx_simulate_A_single(int Nstate, int *start_state, double *P0, matrix<double>& K0, matrix<double>& K1, matrix<double>& K2,
			   matrix<int>& L, matrix<int>& V, matrix<int>& P,
			   double base_amp, double base_var, double *exc_amp, double *exc_var, double vRev,
			   int Nstimulus, int *stim_counts, float **stimuli,
			   int v_signal, double sampling, int Nsample, float *samples, int *states, int *counts, double *amps, double *stds,
			   int seed, int *stop_flag)
{
  timediffs tdiff(20);
  base_generator_type generator( static_cast<boost::mt19937::result_type>(seed) );
  uni_gen_type uni_gen(generator, uni_distribution_type(0.0, 1.0));
  expo_gen_type expo_gen(generator, expo_distribution_type(1.0));
  norm_gen_type norm_gen(generator, norm_distribution_type(0.0, 1.0));

  //tdiff.diff("Asingle");
  // set up stimulus iterator and start state
  stim_intervals stim(Nsample, Nstimulus, stim_counts, stimuli);
  qubx_simulate_setup_P0(1, Nstate, start_state, P0, K0, K1, K2, L, V, P, stim.values, uni_gen);

  //tdiff.diff("c");
  // output state amps, stds for initial stimulus levels
  float dV = 1.0;
  if ( amps ) {
    if ( v_signal ) {
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));
      for (int i=0; i<Nstate; ++i) {
	amps[i] = base_amp + dV * exc_amp[i];
	stds[i] = sqrt(base_var + dV*dV*exc_var[i]);
      }
    } else {
      for (int i=0; i<Nstate; ++i) {
	amps[i] = base_amp + exc_amp[i];
	stds[i] = sqrt(base_var + exc_var[i]);
      }
    }
  }
  
  // set up mechanism and rate constants:
  qubx_simulate_A_single_channel chan(*start_state);
  qubx_simulate_A_kinetics kin(Nstate, K0, K1, K2, L, V, P, sampling);
  
  int at = 0; // output sample index
  int evt_at = 0; // output dwell index

  for (; stim; ++stim) { // for each "count" samples of constant stimulus "values":
    //tdiff.diff("d");
    // update rate constants
    kin.setup_A(stim.values);
    if ( v_signal )
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));

    for (int i=0; i<stim.count; ++i) {
      chan.next_sample(kin.A, uni_gen, Nstate);
      // write gaussian data
      float amp, std;
      if ( v_signal ) {
	amp = (float) (dV * exc_amp[chan.st]);
	std = (float) (sqrt(base_var + dV*dV*exc_var[chan.st]));
      } else {
	amp = (float) exc_amp[chan.st];
	std = (float) sqrt(base_var + exc_var[chan.st]);
      }
      samples[at++] += (float) (base_amp + amp + norm_gen()*std);
      // write event if complete
      if ( states && (chan.next_st != chan.st) ) {
	states[evt_at] = chan.st;
	counts[evt_at] = chan.rep_count;
	++evt_at;
      }
      if ( *stop_flag )
	return evt_at;
    }
  }
  if ( states && (chan.st == chan.next_st) ) { // incomplete event
    states[evt_at] = chan.st;
    counts[evt_at] = chan.rep_count;
    ++evt_at;
  }
  *start_state = chan.next_st;

  //cerr << tdiff << endl;
  return evt_at;
}

int qubx_simulate_A_multi(int Nchannel, int Nstate, int *start_states, double *P0, matrix<double>& K0, matrix<double>& K1, matrix<double>& K2,
			  matrix<int>& L, matrix<int>& V, matrix<int>& P,
			  double base_amp, double base_var, double *exc_amp, double *exc_var, double vRev,
			  int Nstimulus, int *stim_counts, float **stimuli,
			  int v_signal, double sampling, int Nsample, float *samples,
			  QUBFAST_VAR_NOT_USED int *states, QUBFAST_VAR_NOT_USED int *counts, QUBFAST_VAR_NOT_USED double *amps, QUBFAST_VAR_NOT_USED double *stds,
			  int seed, int *stop_flag)
{
  //timediffs tdiff(20);
  base_generator_type generator( static_cast<boost::mt19937::result_type>(seed) );
  uni_gen_type uni_gen(generator, uni_distribution_type(0.0, 1.0));
  expo_gen_type expo_gen(generator, expo_distribution_type(1.0));
  norm_gen_type norm_gen(generator, norm_distribution_type(0.0, 1.0));

  //tdiff.diff("Amulti");
  // set up stimulus iterator
  stim_intervals stim(Nsample, Nstimulus, stim_counts, stimuli);
  
  //tdiff.diff("b");
  qubx_simulate_setup_P0(Nchannel, Nstate, start_states, P0, K0, K1, K2, L, V, P, stim.values, uni_gen);
  
  //tdiff.diff("c");
  // set up mechanism and rate constants:
  std::vector<qubx_simulate_A_single_channel> chans;
  for (int i=0; i<Nchannel; ++i)
    chans.push_back(qubx_simulate_A_single_channel(start_states[i]));
  
  qubx_simulate_A_kinetics kin(Nstate, K0, K1, K2, L, V, P, sampling);

  int at = 0; // output sample index
  int evt_at = 0; // output dwell index

  for (; stim; ++stim) { // for each "count" samples of constant stimulus "values":
    //tdiff.diff("d");
    // update rate constants
    kin.setup_A(stim.values);
    float dV = 1.0;
    if ( v_signal )
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));

    for (int i=0; i<stim.count; ++i) {
      double amp = base_amp;
      double variance = base_var;
      for (int c=0; c<Nchannel; ++c) {
	chans[c].next_sample(kin.A, uni_gen, Nstate);
	if ( v_signal ) {
	  amp += dV * exc_amp[chans[c].st];
	  variance += dV*dV*exc_var[chans[c].st];
	} else {
	  amp += exc_amp[chans[c].st];
	  variance += exc_var[chans[c].st];
	}
      }
      samples[at++] += (float) (amp + norm_gen()*sqrt(variance));
      // write event if complete
      //if ( chan.next_st != chan.st ) {
      //	states[evt_at] = chan.st;
      //	counts[evt_at] = chan.rep_count;
      //	++evt_at;
      //}
      if ( *stop_flag )
	return evt_at;
    }
  }
  //if ( chan.st == chan.next_st ) { // incomplete event
  //  states[evt_at] = chan.st;
  //  counts[evt_at] = chan.rep_count;
  //  ++evt_at;
  //}
  for (int i=0; i<Nchannel; ++i)
    start_states[i] = chans[i].next_st;

  // output magic multi-model class amp and std
  
  //cerr << tdiff << endl;
  return evt_at;
}


    

int qubx_simulate_Q_single(int Nstate, int *start_state, double *P0, matrix<double>& K0, matrix<double>& K1, matrix<double>& K2,
			   matrix<int>& L, matrix<int>& V, matrix<int>& P,
			   double base_amp, double base_var, double *exc_amp, double *exc_var, double vRev,
			   int Nstimulus, int *stim_counts, float **stimuli,
			   int v_signal, double sampling, int Nsample, float *samples, int *states, int *counts, double *amps, double *stds,
			   int seed, int *stop_flag)
{
  //timediffs tdiff(20);
  base_generator_type generator( static_cast<boost::mt19937::result_type>(seed) );
  uni_gen_type uni_gen(generator, uni_distribution_type(0.0, 1.0));
  expo_gen_type expo_gen(generator, expo_distribution_type(1.0));
  norm_gen_type norm_gen(generator, norm_distribution_type(0.0, 1.0));

  //tdiff.diff("Qsingle");
  // set up stimulus iterator and start state
  stim_intervals stim(Nsample, Nstimulus, stim_counts, stimuli);
  qubx_simulate_setup_P0(1, Nstate, start_state, P0, K0, K1, K2, L, V, P, stim.values, uni_gen);

  //tdiff.diff("c");
  // output state amps, stds for initial stimulus levels
  float dV = 1.0;
  if ( amps ) {
    if ( v_signal ) {
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));
      for (int i=0; i<Nstate; ++i) {
	amps[i] = base_amp + dV * exc_amp[i];
	stds[i] = sqrt(base_var + dV*dV*exc_var[i]);
      }
    } else {
      for (int i=0; i<Nstate; ++i) {
	amps[i] = base_amp + exc_amp[i];
	stds[i] = sqrt(base_var + exc_var[i]);
      }
    }
  }
  
  // set up mechanism and rate constants:
  qubx_simulate_Q_single_channel chan(*start_state);
  matrix<double> Q;
  
  int at = 0; // output sample index
  int evt_at = 0; // output dwell index
  double sam = 0.0; // fractional samples
  double maxdur = 0.0; // seconds remaining before stimulus changes or segment ends

  for (; stim; ++stim) { // for each "count" samples of constant stimulus "values":
    //tdiff.diff("d");
    // update rate constants
    BuildQ_p(Nstate, Q, K0, K1, K2, L, V, P, stim.values);
    maxdur = sampling * stim.count;
    if ( v_signal )
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));

    while ( maxdur > 0.0 ) {
      chan.next_event(Q, expo_gen, uni_gen, maxdur, Nstate);
      maxdur -= chan.dur;
      sam += chan.dur / sampling;
      // chan.dur is not an exact multiple of sampling.  Sometimes too short.
      // See if it (cumulatively) crosses the threshold of one sample:
      int evt_n = (int) (sam + 0.5);
      if ( evt_n ) {           // at least one sample long:
	sam -= evt_n;	            // subtract from fractional accumulator
	// write gaussian data
	float amp, std;
	if ( v_signal ) {
	  amp = (float) (dV * exc_amp[chan.st]);
	  std = (float) (sqrt(base_var + dV*dV*exc_var[chan.st]));
	} else {
	  amp = (float) exc_amp[chan.st];
	  std = (float) sqrt(base_var + exc_var[chan.st]);
	}
	for (int i=0; i<evt_n; ++i)
	  samples[at++] += (float) (base_amp + amp + norm_gen()*std);
	if ( states ) {
	  states[evt_at] = chan.st;
	  counts[evt_at] = evt_n;
	  ++evt_at;
	}
	if ( *stop_flag )
	  return evt_at;
      }
    }
  }
  *start_state = chan.next_st;

  //cerr << tdiff << endl;
  return evt_at;
}

int qubx_simulate_Q_multi(int Nchannel, int Nstate, int *start_states, double *P0, matrix<double>& K0, matrix<double>& K1, matrix<double>& K2,
			  matrix<int>& L, matrix<int>& V, matrix<int>& P,
			  double base_amp, double base_var, double *exc_amp, double *exc_var, double vRev,
			  int Nstimulus, int *stim_counts, float **stimuli,
			  int v_signal, double sampling, int Nsample, float *samples,
			  QUBFAST_VAR_NOT_USED int *states, QUBFAST_VAR_NOT_USED int *counts, QUBFAST_VAR_NOT_USED double *amps, QUBFAST_VAR_NOT_USED double *stds,
			  int seed, int *stop_flag)
{
  //timediffs tdiff(20);
  base_generator_type generator( static_cast<boost::mt19937::result_type>(seed) );
  uni_gen_type uni_gen(generator, uni_distribution_type(0.0, 1.0));
  expo_gen_type expo_gen(generator, expo_distribution_type(1.0));
  norm_gen_type norm_gen(generator, norm_distribution_type(0.0, 1.0));

  //tdiff.diff("Qmulti");
  // set up stimulus iterator
  stim_intervals stim(Nsample, Nstimulus, stim_counts, stimuli);
  
  //tdiff.diff("c");
  qubx_simulate_setup_P0(Nchannel, Nstate, start_states, P0, K0, K1, K2, L, V, P, stim.values, uni_gen);
  
  //tdiff.diff("c");
  // set up mechanism and rate constants:
  std::vector<qubx_simulate_Q_single_channel> chans;
  std::vector<int> remain; // number of samples remaining in the current dwell
  std::vector<double> sam; // fractional samples accumulated from prior dwells
  for (int i=0; i<Nchannel; ++i) {
    chans.push_back(qubx_simulate_Q_single_channel(start_states[i]));
    remain.push_back(0); // need to get next event (no samples remaining)
    sam.push_back(0.0);  // no fractional samples accumulated yet
  }
  matrix<double> Q;
  int at = 0; // output sample index
  int evt_at = 0; // output dwell index
  double maxdur = 0.0;

  for (; stim; ++stim) { // for each "count" samples of constant stimulus "values":
    //tdiff.diff("d");
    // update rate constants
    BuildQ_p(Nstate, Q, K0, K1, K2, L, V, P, stim.values);
    maxdur = sampling * stim.count;
    float dV = 1.0;
    if ( v_signal )
      dV = (float) (1e-3 * (stim.values[v_signal] - vRev));

    for (int s=0; s<stim.count; ++s, maxdur-=sampling) {
      double amp = base_amp;
      double variance = base_var;
      for (int c=0; c<Nchannel; ++c) {
	if ( ! remain[c] ) { // this channel's dwell is done:
	  while ( ! remain[c] ) { // accumulate events until it crosses the threshold of one sample:
	    chans[c].next_event(Q, expo_gen, uni_gen, maxdur, Nstate);
	    sam[c] += chans[c].dur / sampling;
	    remain[c] = (int) (sam[c] + 0.5);
	    sam[c] -= remain[c];
	  }
	}
	if ( v_signal ) {
	  amp += dV * exc_amp[chans[c].st];
	  variance += dV*dV*exc_var[chans[c].st];
	} else {
	  amp += exc_amp[chans[c].st];
	  variance += exc_var[chans[c].st];
	}
	--remain[c];
      }
      samples[at++] += (float) (amp + norm_gen()*sqrt(variance));
      // write event if complete
      //if ( chan.next_st != chan.st ) {
      //	states[evt_at] = chan.st;
      //	counts[evt_at] = chan.rep_count;
      //	++evt_at;
      //}
      if ( *stop_flag )
	return evt_at;
    }
  }
  //if ( chan.st == chan.next_st ) { // incomplete event
  //  states[evt_at] = chan.st;
  //  counts[evt_at] = chan.rep_count;
  //  ++evt_at;
  //}
  for (int i=0; i<Nchannel; ++i)
    start_states[i] = chans[i].next_st;

  // output magic multi-model class amp and std
  
  //cerr << tdiff << endl;
  return evt_at;
}


int qubx_simulate_segment(int prefer_Q, int Nchannel, int Nstate, int *start_states, double *P0, matrix<double>& K0, matrix<double>& K1, matrix<double>& K2,
			  matrix<int>& L, matrix<int>& V, matrix<int>& P,
			  double base_amp, double base_var, double *exc_amp, double *exc_var, double vRev,
			  int Nstimulus, int *stim_counts, float **stimuli,
			  int v_signal, double sampling, int Nsample, float *samples, int *states, int *counts, double *amps, double *stds,
			  int seed, int *stop_flag)
{
  if ( prefer_Q ) {
    if ( Nchannel == 1 ) {
      return qubx_simulate_Q_single(Nstate, start_states, P0, K0, K1, K2, L, V, P, base_amp, base_var, exc_amp, exc_var, vRev,
				    Nstimulus, stim_counts, stimuli, v_signal, sampling, Nsample, samples, states, counts, amps, stds, seed, stop_flag);
    } else {
      return qubx_simulate_Q_multi(Nchannel, Nstate, start_states, P0, K0, K1, K2, L, V, P, base_amp, base_var, exc_amp, exc_var, vRev,
				  Nstimulus, stim_counts, stimuli, v_signal, sampling, Nsample, samples, states, counts, amps, stds, seed, stop_flag);
    }
  } else {
    if ( Nchannel == 1 ) {
      return qubx_simulate_A_single(Nstate, start_states, P0, K0, K1, K2, L, V, P, base_amp, base_var, exc_amp, exc_var, vRev,
				    Nstimulus, stim_counts, stimuli, v_signal, sampling, Nsample, samples, states, counts, amps, stds, seed, stop_flag);
    } else {
      return qubx_simulate_A_multi(Nchannel, Nstate, start_states, P0, K0, K1, K2, L, V, P, base_amp, base_var, exc_amp, exc_var, vRev,
				  Nstimulus, stim_counts, stimuli, v_signal, sampling, Nsample, samples, states, counts, amps, stds, seed, stop_flag);
    }
  }
}


bool qubx_simulate_prefer_Q(qubx_model *model, double sampling, int Nsample, int Nstimulus, int *stim_counts, float **stimuli)
{
  qubx_model_storage *mst = (qubx_model_storage *) model->storage;
  stim_intervals stim(Nsample, Nstimulus, stim_counts, stimuli);
  matrix<double> Q;
  double sampling_Hz = 1.0 / sampling;
  
  for (; stim; ++stim) { // for each "count" samples of constant stimulus "values":
    BuildQ_p(model->Nstate, Q, mst->K0.m, mst->K1.m, mst->K2.m, mst->L.m, mst->V.m, mst->P.m, stim.values);
    for (int i=0; i<model->Nstate; ++i) {
      if ( (- Q(i, i)) > (QUBX_SIMULATE_A_THRESHOLD * sampling_Hz) )
	return false;
    }
  }
  return true;
}



extern "C" QUBFAST_API int    qubx_simulate1(qubx_model *model, int *start_states, int add_baseline, int use_Peq, int v_signal,
					    double sampling, int Nstimulus, int *stim_counts, float **stimuli,
					    int Nsample, float *samples, int *states, int *counts, double *amps, double *stds,
					    int force_Q, int seed)
{
  qubx_model_storage *mst = (qubx_model_storage *) model->storage;

  // copy into boost::matrix 
  int Nstate = model->Nstate;
  double base_amp, base_var;
  std::vector<double> exc_amp, exc_var;
  if ( model->IeqFv ) {
    base_amp = model->amp0;
    base_var = model->var0;
    for (int i=0; i<Nstate; ++i) {
      exc_amp.push_back(model->amp[model->clazz[i]]);
      exc_var.push_back(model->var[model->clazz[i]]);
    }
  } else {
    base_amp = model->amp[0];
    base_var = model->var[0];
    for (int i=0; i<Nstate; ++i) {
      exc_amp.push_back(model->amp[model->clazz[i]] - model->amp[0]);
      exc_var.push_back(max(0.0, model->var[model->clazz[i]] - model->var[0]));
    }
  }
  if ( ! add_baseline ) {
    base_amp = base_var = 0.0;
  }

  int prefer_Q = force_Q ? 1 : (qubx_simulate_prefer_Q(model, sampling, Nsample, Nstimulus, stim_counts, stimuli) ? 1 : 0);

  int Ndwell = qubx_simulate_segment(prefer_Q, model->Nchannel,
				     Nstate, &start_states[0], use_Peq ? NULL : model->P0, mst->K0.m, mst->K1.m, mst->K2.m, mst->L.m, mst->V.m, mst->P.m,
				     base_amp, base_var, &exc_amp[0], &exc_var[0], 
				     model->Vrev, Nstimulus, stim_counts, stimuli,
				     v_signal, sampling, Nsample, samples,
				     states, counts, amps, stds, seed, &model->stop_flag);
  return Ndwell;
}

#ifndef EMSCRIPTEN
extern "C" QUBFAST_API int    qubx_simulate(qubx_model *model, int *start_states, int add_baseline, int use_Peq, int v_signal,
				 double sampling, int Nstimulus, int Nseg, int **stim_counts, float ***stimuli,
				 int *Nsample, float **samples, int *ndwt, int **states, int **counts, double **amps, double **stds,
				 int force_Q, StatusCallback pctCB, void *pctObj)
{
  //print_timestamp("-----enter qubx_simulate_Q");
  seed_gen_type seed_gen(g_randgen_qubxsim, seed_distribution_type(-(1<<30), 1<<30));
  int seed = seed_gen();

  qubx_model_storage *mst = (qubx_model_storage *) model->storage;

  // copy into boost::matrix 
  int Nstate = model->Nstate;
  
  double base_amp, base_var;
  std::vector<double> exc_amp, exc_var;
  if ( model->IeqFv ) {
    base_amp = model->amp0;
    base_var = model->var0;
    for (int i=0; i<Nstate; ++i) {
      exc_amp.push_back(model->amp[model->clazz[i]]);
      exc_var.push_back(model->var[model->clazz[i]]);
    }
  } else {
    base_amp = model->amp[0];
    base_var = model->var[0];
    for (int i=0; i<Nstate; ++i) {
      exc_amp.push_back(model->amp[model->clazz[i]] - model->amp[0]);
      exc_var.push_back(max(0.0, model->var[model->clazz[i]] - model->var[0]));
    }
  }
  if ( ! add_baseline ) {
    base_amp = base_var = 0.0;
  }

  int prefer_Q_count = 0;
  std::vector<int> prefer_Q(Nseg);
  if ( force_Q ) {
    for (int i=0; i<Nseg; ++i)
      prefer_Q[i] = 1;
    prefer_Q_count = Nseg;
  } else {
    #pragma omp parallel for reduction(+:prefer_Q_count)
    for (int i=0; i<Nseg; ++i) {
      prefer_Q_count += (prefer_Q[i] = (qubx_simulate_prefer_Q(model, sampling, Nsample[i], Nstimulus, stim_counts[i], stimuli[i]) ? 1 : 0));
    }
  }
  
  if ( Nseg > 1 ) {
    #pragma omp parallel
    {
      std::vector<int> thread_start_states(model->Nchannel);
      
      #pragma omp for schedule(dynamic)
      for (int i=0; i<Nseg; ++i) {
	memcpy(& thread_start_states[0], start_states, model->Nchannel*sizeof(int));
	int nd = qubx_simulate_segment(prefer_Q[i], model->Nchannel,
				       Nstate, &thread_start_states[0], use_Peq ? NULL : model->P0, mst->K0.m, mst->K1.m, mst->K2.m, mst->L.m, mst->V.m, mst->P.m,
				       base_amp, base_var, &exc_amp[0], &exc_var[0], 
				       model->Vrev, Nstimulus, stim_counts[i], stimuli[i],
				       v_signal, sampling, Nsample[i], samples[i],
				       states ? states[i] : NULL, counts ? counts[i] : NULL, amps ? amps[i] : NULL, stds ? stds[i] : NULL,
				       seed+i, &model->stop_flag);
	if ( ndwt )
	  ndwt[i] = nd;
	if ( 0 == omp_get_thread_num() ) {
	  if ( pctCB(pctObj, i*1.0/Nseg) )
	    model->stop_flag = 1;
	}
      }
      #pragma omp single
      {
	memcpy(start_states, & thread_start_states[0], model->Nchannel*sizeof(int));
      }
    }
  } else if ( (model->Nchannel >= QUBX_SIMULATE_SPLIT_THRESHOLD) && (Nsample[0] < QUBX_SIMULATE_SPLIT_MAXPOINTS) ) {
    std::vector< std::vector<float> > thread_samples;
    #pragma omp parallel
    {
      int iThread = omp_get_thread_num();
      int nThread = omp_get_num_threads();
      #pragma omp single
      {
	thread_samples.resize(nThread);
	for (int i=0; i<nThread; ++i)
	  thread_samples[i].resize(Nsample[0]);
      }
      int nChannel = model->Nchannel / nThread + ((iThread < (model->Nchannel % nThread)) ? 1 : 0);
      int iChannel = iThread*(model->Nchannel / nThread) + min(iThread, (model->Nchannel % nThread));
      qubx_simulate_segment(prefer_Q[0], nChannel,
			    Nstate, start_states+iChannel, use_Peq ? NULL : model->P0, mst->K0.m, mst->K1.m, mst->K2.m, mst->L.m, mst->V.m, mst->P.m,
			    base_amp, base_var, &exc_amp[0], &exc_var[0], 
			    model->Vrev, Nstimulus, stim_counts[0], stimuli[0],
			    v_signal, sampling, Nsample[0], & thread_samples[iThread][0],
			    states ? states[0] : NULL, counts ? counts[0] : NULL, amps ? amps[0] : NULL, stds ? stds[0] : NULL,
			    seed+iThread, &model->stop_flag);
      #pragma omp barrier
      int nData = Nsample[0] / nThread + ((iThread < (Nsample[0] % nThread)) ? 1 : 0);
      int iData = iThread*(Nsample[0] / nThread) + min(iThread, (Nsample[0] % nThread));
      for (int i=iData+nData-1; i>=iData; --i) {
	double sum = 0.0;
	for (int t=0; t<nThread; ++t)
	  sum += thread_samples[t][i];
	samples[0][i] += (float) sum;
      }
    }
    if ( ndwt )
      ndwt[0] = 0;
  } else {
    int nd = qubx_simulate_segment(prefer_Q[0], model->Nchannel,
				   Nstate, start_states, use_Peq ? NULL : model->P0, mst->K0.m, mst->K1.m, mst->K2.m, mst->L.m, mst->V.m, mst->P.m,
				   base_amp, base_var, &exc_amp[0], &exc_var[0], 
				   model->Vrev, Nstimulus, stim_counts[0], stimuli[0],
				   v_signal, sampling, Nsample[0], samples[0],
				   states ? states[0] : NULL, counts ? counts[0] : NULL, amps ? amps[0] : NULL, stds ? stds[0] : NULL,
				   seed, &model->stop_flag);
    if ( ndwt )
      ndwt[0] = nd;
  }
  //print_timestamp("-----");

  return 0;
}
#endif

extern "C" QUBFAST_API int  qubx_simulate_make_idealized_flc(int *ff, int *ll, int *cc, int N, int *clazz, int *states, int *st_counts, int at)
{
  for (int i=N; i; --i, ++ff, ++ll, ++cc, ++states, ++st_counts) {
    *ff = at;
    at += *st_counts;
    *ll = at - 1;
    *cc = clazz[*states];
  }
  
  return N; // TODO: condense same-class events and return smaller N
}


void qubx_simulate_baseline(int Nsample, float *samples, double sampling,
			    int baselineFluct, int baselineFluctN, double baselineFluctLifetime, double baselineFluctMaxAmp,
			    int baselineDrift, double baselineDriftSlope, double baselineDriftIntercept,
			    int baselinePeriodic, double baselinePeriodicFreq, double baselinePeriodicPhase, double baselinePeriodicAmp,
			    int baselineJumps, double baselineJumpsAmpStd,
			    double baselineJumpsFastW, double baselineJumpsFast, double baselineJumpsSlow, double baselineJumpsLifetime,
			    int baselineLGM, int baselineLGM2State, double baselineLGMProcStd,
			    double baselineLGMX0, double baselineLGMX0p, int seed)
{
  base_generator_type generator( static_cast<boost::mt19937::result_type>(seed) );
  uni_gen_type uni_gen(generator, uni_distribution_type(0.0, 1.0));
  expo_gen_type expo_gen(generator, expo_distribution_type(1.0));
  norm_gen_type norm_gen(generator, norm_distribution_type(0.0, 1.0));

  int baselineFluctStateCount = 0;
  double bufDuration = Nsample * sampling;

  //Baseline Fluctuations
  if ( baselineFluct ) {
    std::vector<int> flucBuffer(Nsample);
    int closeCount = abs((baselineFluctN - abs(baselineFluctStateCount)) / 2);
    if ( baselineFluctStateCount < 0 )
      closeCount = - closeCount;

    for (int c=1; c<=baselineFluctN; ++c) {
      //Event generator loop
      int baseCurPoints      = 0;
      double baseCurDuration = 0;
      double baseDuration    = 0;
      int baseState, basePoints;
      if ( c > closeCount )
	baseState = 1;
      else
	baseState = -1;
      
      do {
	//Generate a baseline event at least Sampling long
	while ( baseDuration < sampling ) {
	  baseState = -baseState;
	  baseDuration = baseDuration + expo_gen() * baselineFluctLifetime;
	}
	
	if ( (baseCurDuration + baseDuration) > bufDuration )
	  basePoints   = Nsample - baseCurPoints;
	else
	  basePoints   = int(0.5 + (baseDuration / sampling));
	if ( basePoints == 0 )
	  basePoints = 1;
	baseDuration = basePoints * sampling;
	
	for (int p=baseCurPoints; p < baseCurPoints+basePoints; ++p)
	  flucBuffer[p] = flucBuffer[p] + baseState;
	
	//The remaining duration
	baseDuration = baseDuration - sampling * basePoints;
	
	baseCurPoints   = baseCurPoints + basePoints;
	baseCurDuration = baseCurPoints * sampling;
	
      } while ( baseCurPoints < Nsample );
      
    }//of the ith / N particle
    
    double t = baselineFluctMaxAmp / baselineFluctN;
    for (int p=0; p<Nsample; ++p)
      samples[p] += (float) (t * flucBuffer[p]);
    
    baselineFluctStateCount = flucBuffer[Nsample - 1]; // to be carried over globally?
  }//of BaselineFluct

  //Baseline Drift
  if ( baselineDrift ) {
    for (int p=0; p<Nsample; ++p)
      samples[p] += (float) (baselineDriftSlope * p * sampling + baselineDriftIntercept);
  }
  
  //Baseline Periodic
  if ( baselinePeriodic ) {
    double Frq   = baselinePeriodicFreq * 2.0 * M_PI;
    double Phs   = baselinePeriodicPhase * M_PI / 180.0;
    
    for (int p=0; p<Nsample; ++p)
      samples[p] += (float) (baselinePeriodicAmp * sin(Frq * p * sampling + Phs));
  }

  //Baseline Jumps
  if ( baselineJumps ) {
    int baselineJumpsStateStart = 0;
    double baselineJumpsStateDur = expo_gen() * baselineJumpsLifetime;
    double baselineJumpsStateAmpF = 0.0;
    double baselineJumpsStateAmpL = 0.0;
    double baselineJumpsStateAmpC = 0.0;

    //Event generator loop
    int baseCurPoints      = 0;
    double baseCurDuration = 0;
    double baseDuration    = 0.0;
    int basePoints;

    do {
      //Generate a baseline event at least Sampling long
      while ( baseDuration < sampling ) {
	//Generate new jump
	baselineJumpsStateAmpF  = (2 * int(0.5 + uni_gen()) - 1) * (baselineJumpsStateAmpC + baselineJumpsAmpStd * norm_gen());
	baselineJumpsStateAmpL  = baselineJumpsStateAmpC;
	
	baselineJumpsStateStart = baseCurPoints;
	baselineJumpsStateDur   = expo_gen() * baselineJumpsLifetime;
	baseDuration = baseDuration + baselineJumpsStateDur;
      }

      if ( (baseCurDuration + baseDuration) > bufDuration )
	basePoints   = Nsample - baseCurPoints;
      else
	basePoints   = int(0.5 + (baseDuration / sampling));
      if ( basePoints == 0 )
	basePoints = 1;
      baseDuration = basePoints * sampling;
      
      double e = baselineJumpsStateAmpC;
      for (int p=baseCurPoints; p<baseCurPoints+basePoints; ++p) {
	e = baselineJumpsFastW * exp(-(p - baselineJumpsStateStart) * sampling / baselineJumpsFast/*sec*/);
	e += (1.0 - baselineJumpsFastW) * exp(-(p - baselineJumpsStateStart) * sampling / baselineJumpsSlow/*sec*/);
	e = baselineJumpsStateAmpL + baselineJumpsStateAmpF * e;
	samples[p] += (float) e;
      }
      baselineJumpsStateAmpC = e;

      //The remaining duration
      baseDuration = baseDuration - sampling * basePoints;
      baselineJumpsStateDur = baseDuration;

      baseCurPoints   = baseCurPoints + basePoints;
      baseCurDuration = baseCurPoints * sampling;
    } while ( baseCurPoints < Nsample );
  }//of Baseline Jumps

  //Baseline LGM
  if ( baselineLGM ) {
    double baselineLGMState[2];
    baselineLGMState[0] = baselineLGMX0;
    baselineLGMState[1] = baselineLGMX0p;
    
    if ( baselineLGM2State ) {
      //2 states: x and x'
      for (int p=0; p<Nsample; ++p ) {
	samples[p] += (float) baselineLGMState[0];
	baselineLGMState[0] += baselineLGMState[1];
	baselineLGMState[1] += baselineLGMProcStd * norm_gen();
      }
    } else {
      //one state: x
      for (int p=0; p<Nsample; ++p ) {
	samples[p] += (float) baselineLGMState[0];
	baselineLGMState[0] += baselineLGMProcStd * norm_gen();
      }
    }
  }
}

extern "C" QUBFAST_API void qubx_simulate_baselines(int Nsegment, int *Nsamples, float **sample_arrays, double sampling,
						    int baselineFluct, int baselineFluctN, double baselineFluctLifetime, double baselineFluctMaxAmp,
						    int baselineDrift, double baselineDriftSlope, double baselineDriftIntercept,
						    int baselinePeriodic, double baselinePeriodicFreq, double baselinePeriodicPhase, double baselinePeriodicAmp,
						    int baselineJumps, double baselineJumpsAmpStd,
						    double baselineJumpsFastW, double baselineJumpsFast, double baselineJumpsSlow, double baselineJumpsLifetime,
						    int baselineLGM, int baselineLGM2State, double baselineLGMProcStd,
						    double baselineLGMX0, double baselineLGMX0p)
{
  seed_gen_type seed_gen(g_randgen_qubxsim, seed_distribution_type(-(1<<30), 1<<30));
  int seed = seed_gen();

  if ( Nsegment > 1 ) {
#pragma omp parallel for schedule(dynamic)
    for (int i=0; i<Nsegment; ++i)
      qubx_simulate_baseline(Nsamples[i], sample_arrays[i], sampling,
			     baselineFluct, baselineFluctN, baselineFluctLifetime, baselineFluctMaxAmp,
			     baselineDrift, baselineDriftSlope, baselineDriftIntercept,
			     baselinePeriodic, baselinePeriodicFreq, baselinePeriodicPhase, baselinePeriodicAmp,
			     baselineJumps, baselineJumpsAmpStd,
			     baselineJumpsFastW, baselineJumpsFast, baselineJumpsSlow, baselineJumpsLifetime,
			     baselineLGM, baselineLGM2State, baselineLGMProcStd, baselineLGMX0, baselineLGMX0p, seed+i);
  } else if ( Nsegment > 0 ) {
    qubx_simulate_baseline(Nsamples[0], sample_arrays[0], sampling,
			   baselineFluct, baselineFluctN, baselineFluctLifetime, baselineFluctMaxAmp,
			   baselineDrift, baselineDriftSlope, baselineDriftIntercept,
			   baselinePeriodic, baselinePeriodicFreq, baselinePeriodicPhase, baselinePeriodicAmp,
			   baselineJumps, baselineJumpsAmpStd,
			   baselineJumpsFastW, baselineJumpsFast, baselineJumpsSlow, baselineJumpsLifetime,
			   baselineLGM, baselineLGM2State, baselineLGMProcStd, baselineLGMX0, baselineLGMX0p, seed);
  }
}


extern "C" QUBFAST_API void qubx_simulate_stimulus_steps(int Nsample, float *samples, double sampling,
							 int Nstep, float *levels, float *durations)
{
  float y = 0.0f;
  double x = 0.0; // samples
  int j = 0;

  for (int i=0; i<Nstep; ++i) {
    y = levels[i];
    x += (durations[i] / sampling);
    int jStop = min(Nsample, (int) ceil(x));
    for (; j<jStop; ++j)
      samples[j] = y;
  }
  for (; j<Nsample; ++j)
    samples[j] = y;
}

extern "C" QUBFAST_API void qubx_simulate_stimulus_ramps(int Nsample, float *samples, double sampling,
							 int Nstep, float *levels, float *durations)
{
  double y0 = 0.0;
  float y1 = 0.0;
  double x = 0.0;
  int j = 0;

  for (int i=0; i<Nstep; ++i) {
    y0 = y1;
    y1 = levels[i];
    double dur_samples = durations[i] / sampling;
    double m = (y1 - y0) / dur_samples;
    x += dur_samples;
    int jStop = min(Nsample, (int) ceil(x));
    for (int k=0; j<jStop; ++j, ++k)
      samples[j] = (float) (m*k + y0);
  }
  for (; j<Nsample; ++j)
    samples[j] = y1;
}


